#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/resource.h>


#include "imapfilter.h"


extern unsigned int options;
extern uid_t ruid, euid;

static secmem_t *smem = NULL;	/* First node of secure memory linked list. */


/*
 * A malloc() that checks the results and dies in case of error.
 */
void *xmalloc(size_t size)
{
    void *ptr;

    ptr = (void *)malloc(size);

    if (!ptr)
	fatal(ERROR_MEMORY_ALLOCATION,
	      "imapfilter: allocating memory; %s\n", strerror(errno));

    return ptr;
}


/*
 * A realloc() that checks the results and dies in case of error.
 */
void *xrealloc(void *ptr, size_t size)
{
    ptr = (void *)realloc(ptr, size);

    if (!ptr)
	fatal(ERROR_MEMORY_ALLOCATION,
	      "imapfilter: allocating memory; %s\n", strerror(errno));

    return ptr;
}


/*
 * A free() that dies if fed with NULL pointer.
 */
void xfree(void *ptr)
{
    if (!ptr)
	fatal(ERROR_MEMORY_ALLOCATION,
	      "imapfilter: NULL pointer given as argument");
    free(ptr);
}


/*
 * A strdup() that checks the results and dies in case of error.
 */
char *xstrdup(const char *s)
{
    char *cp;

    cp = strdup(s);

    if (!cp)
	fatal(ERROR_MEMORY_ALLOCATION,
	      "imapfilter: allocating memory; %s\n", strerror(errno));

    return cp;
}


/*
 * Secure memory malloc().  Locks memory and keeps information about the
 * chunk that was allocated.
 */
void *smalloc(size_t size)
{
    int r;
    void *ptr;
    static int w = 0;
    secmem_t *node;

    ptr = xmalloc(size);

    seteuid(euid);		/* Gain root privileges. */
    r = mlock(ptr, size);
    seteuid(ruid);		/* Drop root privileges. */

    if (getuid() != geteuid())
	fatal(ERROR_SETUID, "imapfilter: failed to drop privileges\n");

    if (options & OPTION_WARNING && r && !w) {
	error("imapfilter: warning: using insecure memory\n");
	w = 1;
    }
    node = (secmem_t *) xmalloc(sizeof(secmem_t));

    node->buf = ptr;
    node->size = size;
    node->prev = node->next = NULL;

    secmem_append(node);

    return ptr;
}


/*
 * Secure memory realloc().  Resize memory by allocating a new memory chunk
 * and NULL fill old memory, in order to protect sensitive data.
 */
void *srealloc(void *ptr, size_t size)
{
    void *p;
    secmem_t *node;

    if (!(node = (secmem_t *) secmem_find(ptr))) {
	ptr = xrealloc(ptr, size);
	return ptr;
    }
    p = smalloc(size);
    memcpy(p, node->buf, min(node->size, size));

    memset(node->buf, 0, node->size);
    secmem_remove(node);
    xfree(node->buf);
    xfree(node);

    return p;
}


/*
 * Secure memory free().  NULL fill memory before freeing it.
 */
void sfree(void *ptr)
{
    secmem_t *node;

    if (!(node = (secmem_t *) secmem_find(ptr))) {
	xfree(ptr);
	return;
    }
    memset(node->buf, 0, node->size);

    secmem_remove(node);
    xfree(node->buf);
    xfree(node);
}


/*
 * Secure memory strdup().  Uses secure memory allocation.
 */
char *sstrdup(const char *s)
{
    char *p;

    p = (char *)smalloc(strlen(s) + 1);
    xstrncpy(p, s, strlen(s));

    return p;
}


/*
 * Append information about the newly allocated memory buffer.
 */
void secmem_append(secmem_t * node)
{
    secmem_t *pos;
    secmem_t **app;

    app = &smem;
    pos = smem;

    while (pos) {
	node->prev = pos;
	app = &(pos->next);
	pos = pos->next;
    }

    *app = node;
}


/*
 * Find the record of a memory buffer in the secure memory linked list.
 */
secmem_t *secmem_find(void *ptr)
{
    secmem_t *pos;

    pos = smem;

    while (pos && pos->buf != ptr)
	pos = pos->next;

    return pos;
}


/*
 * Remove a record of a secure memory buffer.
 */
void secmem_remove(secmem_t * node)
{
    if (node->prev)
	node->prev->next = node->next;
    if (node->next)
	node->next->prev = node->prev;
}


/*
 * Overwrite/clear all secure memory.
 */
void secmem_clear(void)
{
    secmem_t *p;

    for (p = smem; p; p = p->next)
	sfree(p);
}


/*
 * Lock memory of allocated buffers.
 */
void secmem_lock(void)
{
    secmem_t *p;

    seteuid(euid);		/* Gain root privileges. */
    for (p = smem; p; p = p->next)
	mlock(p->buf, p->size);
    seteuid(ruid);		/* Drop root privileges. */

    if (getuid() != geteuid())
	fatal(ERROR_SETUID, "imapfilter: failed to drop privileges\n");
}


/*
 * Disable core file dumping.
 */
void corefile_disable(void)
{
    struct rlimit rl;

    getrlimit(RLIMIT_CORE, &rl);

    rl.rlim_cur = rl.rlim_max = 0;
    setrlimit(RLIMIT_CORE, &rl);
}
